/** \file error.h
	\brief Hosts declarations for error and exception related functions
	\cond DEV
*/

#ifndef AUDRECORDCLI_AUDRECORD_H
#define AUDRECORDCLI_AUDRECORD_H

#pragma once

#include <boost/type_traits/is_same.hpp>
#include <boost/type_traits/remove_pointer.hpp>

namespace AudRecordCLI
{
	/** Generic exception class that covers failure HRESULTs
		from the native functions that aren't already
		covered by the builtin exception classes
	*/
	public ref class AudRecordException : public Exception
	{
	public:
		AudRecordException(String^ errorText, HRESULT hr);
	};
}

/** Retrieves a message string from a module or the system message table

	\tparam Args Type of the additional arguments, can be va_list or DWORD_PTR
	\param hMod Module to search for the message in, or NULL for the system
	\param messageId Message to lookup
	\param [out] errorText The retrieved error text
	\param args The arguments to insert, no substitution is done if this is NULL
	\returns Nonzero if the message was found, zero otherwise
*/
template<class Args>
BOOL GetModuleMessage(HMODULE hMod, DWORD messageId, String^& errorText, Args* args)
{
	DWORD flags = FORMAT_MESSAGE_ALLOCATE_BUFFER;
	if(!args)
	{
		flags |= FORMAT_MESSAGE_IGNORE_INSERTS;
	}
	// this is a compile time check, its removed by optimization
	else if(boost::is_same<typename boost::remove_pointer<Args>::type, DWORD_PTR>::value)
	{
		flags |= FORMAT_MESSAGE_ARGUMENT_ARRAY;
	}
	flags |= hMod ? FORMAT_MESSAGE_FROM_HMODULE : FORMAT_MESSAGE_FROM_SYSTEM;
	WCHAR* error;
	if(FormatMessageW(
			flags,
			hMod,
			messageId,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			reinterpret_cast<WCHAR*>(&error),
			0, reinterpret_cast<va_list*>(args)
		)
	)
	{
		errorText = gcnew String(error);
		LocalFree(error);
	}
	return !!error;
}

// these aer documented in the cpp file
String^ GetModuleString(DWORD stringId);
BOOL GetModuleMessage(HMODULE hMod, DWORD messageId, String^& errorText);
void ThrowExceptionIfRequired(HRESULT hr);
String^ FindErrorString(HRESULT hr);

/** \endcond */

#endif
