/** \file interfacewrap.cpp
	\brief Implements a basic wrapper around an ITargetSink interface
	that allows it to be passed to the native IAudRecordCLI::AddCustomSink()
	\cond DEV
*/

#include "stdafx.h"
#include "InterfaceWrap.h"
#include "DataBufferWrap.h"

/** Constructs the wrapper around the specified target object

	\param pOrigSink The target to wrap
*/
TargetSinkWrap::TargetSinkWrap(AudRecordCLI::ITargetSink^ pOrigSink)
	: pOrigSink(pOrigSink),
	refCount(1)
{}

/** Returns pointers for implemented interfaces on the object

	Currently supported interfaces are IUnknown and IWMWriterSink
	\param iid The id of the requested interface
	\param [out] ppv A pointer to a pointer that receives the interface pointer or NULL
	\return S_OK on success, E_NOINTERFACE for unsupported interfaces and E_POINTER if ppObj is NULL
*/
STDMETHODIMP TargetSinkWrap::QueryInterface(REFIID iid, PVOID* ppv)
{
	if(!ppv) return E_POINTER;
	if(iid == IID_IWMWriterSink || iid == IID_IUnknown)
	{
		AddRef();
		*ppv = this;
		return S_OK;
	}
	*ppv = NULL;
	return E_NOINTERFACE;
}

/** Increments the objects reference count

	\returns The new reference count
*/
STDMETHODIMP_(ULONG) TargetSinkWrap::AddRef()
{
	return _InterlockedIncrement(&refCount);
}

/** Drecreases the objects reference count

	If the ref count reaches zero, the object is destroyed
	\returns The new reference count
*/
STDMETHODIMP_(ULONG) TargetSinkWrap::Release()
{
	LONG newCount = _InterlockedDecrement(&refCount);
	if(newCount <= 0)
	{
		delete this;
	}
	return static_cast<ULONG>(newCount);
}

/** Called when there's a header to be written.
	
	The native buffer is wrapped into a managed IDataBuffer
	and passed to RaiseOnHeader which fires the event.
	Afterwards, the managed wrapper is deleted. Any exceptions
	arising from raising the event are caught, logged
	and cause the function to return E_FAIL

	\param pHeader A native buffer that contains the header data
	\returns S_OK on success, E_POINTER if pHeader is NULL or E_FAIL
	if raising the managed event throws	an exception
*/
STDMETHODIMP TargetSinkWrap::OnHeader(INSSBuffer* pHeader)
{
	if(!pHeader) return E_POINTER;
	HRESULT hr = S_OK;
	AudRecordCLI::IDataBuffer^ pWrap = WrapNativeBuffer(pHeader);
	try
	{
		pOrigSink->RaiseOnHeader(pWrap);
	}
	catch(System::Exception^ ex)
	{
		pin_ptr<const WCHAR> reason = PtrToStringChars(ex->Message);
		pin_ptr<const WCHAR> stackTrace = PtrToStringChars(ex->StackTrace);
		hr = E_FAIL;
		DBG_TEXT("Caught exception '%S' from managed OnHeader event(s):\n%S", reason, stackTrace);
	}
	delete pWrap;
	return hr;
}

/** Determines whether the wrapped sink wants realtime packet delivery

	The function simply calls the managed IsRealTime and puts the
	returned value in the pointer parameter

	\param pRealTime Pointer that receives the value returned from ITargetSink::IsRealTime()
	\returns S_OK on success, E_POINTER if pRealTime is NULL, or E_FAIL if 
	calling the managed IsRealTime throws an exception
*/
STDMETHODIMP TargetSinkWrap::IsRealTime(BOOL* pRealTime)
{
	if(!pRealTime) return E_POINTER;
	HRESULT hr = S_OK;
	try
	{
		*pRealTime = pOrigSink->IsRealTime();
	}
	catch(System::Exception^ ex)
	{
		pin_ptr<const WCHAR> reason = PtrToStringChars(ex->Message);
		pin_ptr<const WCHAR> stackTrace = PtrToStringChars(ex->StackTrace);
		hr = E_FAIL;
		DBG_TEXT("Caught exception '%S' from managed IsRealTime function:\n%S", reason, stackTrace);
	}
	return hr;
}

/** Allocates a buffer of specified size and returns it to the caller

	First the function calls the managed AllocateDataUnit to create
	a buffer in whichever way it sees fit. If it returns successfully,
	the IDataBuffer interface it returns is wrapped in a ManagedDataBufferWrap
	which is returned to the caller via its INSSBuffer interface

	\param dataSize Number of bytes to allocate
	\param ppBuffer A pointer to an interface on the returned buffer
	\returns S_OK on success, E_POINTER if ppBuffer is NULL, E_OUTOFMEMORY if 
	the managed AllocateDataUnit returns null or wrapping it in a native wrapper
	fails, or E_FAIL if the managed code throws an exception
*/
STDMETHODIMP TargetSinkWrap::AllocateDataUnit(DWORD dataSize, INSSBuffer** ppBuffer)
{
	if(!ppBuffer) return E_POINTER;
	HRESULT hr = S_OK;
	AudRecordCLI::IDataBuffer^ pManBuffer = nullptr;
	try
	{
		pManBuffer = pOrigSink->AllocateDataUnit(dataSize);
		if(!pManBuffer)
		{
			hr = E_OUTOFMEMORY;
		}
	}
	catch(System::Exception^ ex)
	{
		pin_ptr<const WCHAR> reason = PtrToStringChars(ex->Message);
		pin_ptr<const WCHAR> stackTrace = PtrToStringChars(ex->StackTrace);
		hr = E_FAIL;
		DBG_TEXT("Caught exception '%S' from managed AllocateDataUnit event(s):\n%S", reason, stackTrace);
	}
	if(SUCCEEDED(hr))
	{
		if(!(*ppBuffer = WrapManagedBuffer(pManBuffer)))
		{
			hr = E_OUTOFMEMORY;
		}
	}
	return hr;
}

/** Called when there's a buffer of data to be written.
	
	The native buffer is wrapped into a managed IDataBuffer
	and passed to RaiseOnDataUnit which fires the event.
	Afterwards, the managed wrapper is deleted. Any exceptions
	arising from raising the event are caught, logged
	and cause the function to return E_FAIL

	\param pBuffer A native buffer that contains the data to be written
	\returns S_OK on success, E_POINTER is pBuffer is NULL, or E_FAIL
	if raising the managed event throws	an exception
*/
STDMETHODIMP TargetSinkWrap::OnDataUnit(INSSBuffer* pBuffer)
{
	if(!pBuffer) return E_POINTER;
	HRESULT hr = S_OK;
	AudRecordCLI::IDataBuffer^ pManBuffer = WrapNativeBuffer(pBuffer);
	try
	{
		pOrigSink->RaiseOnDataUnit(pManBuffer);
	}
	catch(System::Exception^ ex)
	{
		pin_ptr<const WCHAR> reason = PtrToStringChars(ex->Message);
		pin_ptr<const WCHAR> stackTrace = PtrToStringChars(ex->StackTrace);
		hr = E_FAIL;
		DBG_TEXT("Caught exception '%S' from managed OnDataUnit event(s):\n%S", reason, stackTrace);
	}
	delete pManBuffer;
	return hr;
}

/** Called when writing has finished for this session

	Simply raises the managed event of the same name

	\returns S_OK on success, or E_FAIL if the managed event
	handler throws an exception
*/
STDMETHODIMP TargetSinkWrap::OnEndWriting()
{
	HRESULT hr = S_OK;
	try
	{
		pOrigSink->RaiseOnEndWriting();
	}
	catch(System::Exception^ ex)
	{
		pin_ptr<const WCHAR> reason = PtrToStringChars(ex->Message);
		pin_ptr<const WCHAR> stackTrace = PtrToStringChars(ex->StackTrace);
		hr = E_FAIL;
		DBG_TEXT("Caught exception '%S' from managed OnEndWriting event(s):\n%S", reason, stackTrace);
	}
	return hr;
}

namespace AudRecordCLI
{
	/** Internal function which raises the OnHeader event

		\param pBuffer The data packet to be passed to the event
	*/
	void ITargetSink::RaiseOnHeader(IDataBuffer^ pBuffer)
	{
		OnHeader(this, pBuffer);
	}

	/** Internal function which raises the OnEndWriting event */
	void ITargetSink::RaiseOnEndWriting()
	{
		OnEndWriting(this);
	}

	/** Internal function which raises the OnDataUnit event

		\param pBuffer The data packet containing the data to be passed
	*/
	void ITargetSink::RaiseOnDataUnit(IDataBuffer^ pBuffer)
	{
		OnDataUnit(this, pBuffer);
	}
}

/** \endcond */
