/** \file interfacewrap.h
	\brief Declares elements required for a managed	custom sink 
	to be passed to the IAudRecord interface
*/
#ifndef AUDRECORDCLI_INTERFACEWRAP_H
#define AUDRECORDCLI_INTERFACEWRAP_H

#pragma once

#include "DataBufferWrap.h"

namespace AudRecordCLI
{
	/** The form of an event fired when a data packet is ready for writing

		\param sender The object that sends the event
		\param dataBuffer The buffer containing the data to be written
	*/
	public delegate void OnDataUnitHandler(Object^ sender, IDataBuffer^ dataBuffer);

	/** The form of an event fired when all writing to an ITargetSink has finished.
		If the target hasn't been removed before Recorder.Record() is called again
		it will receive data as normal

		\param sender The object that sends the event
	*/
	public delegate void OnEndWritingHandler(Object^ sender);

	/** The form of an event fired when the ASF file header is ready for writing.
		This can be sent multiple times with different information

		\param sender The object that sends the event
		\param headerBuffer The buffer containing the header data to be written
	*/
	public delegate void OnHeaderHandler(Object^ sender, IDataBuffer^ headerBuffer);

	/** The base interface for a custom target. This is a ref struct
		rather than an ref interface because we need to have the 
		event raising methods as members.
	*/
	public ref struct ITargetSink abstract
	{
		/** Event raised when the ASF header is ready to write
			headerBuffer contains the data. Can be raised
			more than once
		*/
		event OnHeaderHandler^ OnHeader;
		/** Called by Windows Media runtime to ascertain whether the
			target needs data samples to be delivered in real time

			\returns true if the sink needs realtime data, false if it can live without
		*/
		virtual bool IsRealTime() = 0;
		/** Called by Windows Media runtime to ascertain whether the
			target needs data samples to be delivered in real time

			\param dataSize The size of data buffer to allocate
			\returns The new data buffer object
		*/
		virtual IDataBuffer^ AllocateDataUnit(DWORD dataSize) = 0;
		/** Event raised when a buffer has been filled with data and
			is ready to be written in whatever way is desired
		*/
		event OnDataUnitHandler^ OnDataUnit;
		/** Event raised when all writing to a target has finished.
			If the target is still present in the Recorder when
			Record() is called again, it will start receiving calls
			again
		*/
		event OnEndWritingHandler^ OnEndWriting;
	internal:
		void RaiseOnHeader(IDataBuffer^ pBuffer);
		void RaiseOnDataUnit(IDataBuffer^ pBuffer);
		void RaiseOnEndWriting();
	};
}

/** \cond DEV */
/**
	Native wrapper around an ITargetSink interface which
	enables it to be passed to the AddCustomSink method
	exposed by IAudRecord
*/
class TargetSinkWrap : public IWMWriterSink
{
	/** The wrapper around the managed interface that allows
		it to be embedded in a native object
	*/
	gcroot<AudRecordCLI::ITargetSink^> pOrigSink;
	/** Reference count of the wrapper. The object will be
		deleted when it drops to zero
	*/
	LONG refCount;
public:
	TargetSinkWrap(AudRecordCLI::ITargetSink^ pOrigSink);
	// IUnknown
	STDMETHODIMP QueryInterface(REFIID riid, PVOID* ppv);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();
	// IWMWriterSink
	STDMETHODIMP OnHeader(INSSBuffer* pHeader);
	STDMETHODIMP IsRealTime(BOOL* pRealTime);
	STDMETHODIMP AllocateDataUnit(DWORD dataSize, INSSBuffer** ppBuffer);
	STDMETHODIMP OnDataUnit(INSSBuffer* pBuffer);
	STDMETHODIMP OnEndWriting();
};

/** \endcond */

#endif
