/** \file alloc.cpp
	\brief Implements memory management, and the default function used
	if no user defined one have been specified
	\cond DEV
*/

#include "stdafx.h"
#include "alloc.h"

/** Data added to the start of each allocation

	This ensures that data allocated by one allocator
	will be freed by it's partner free, even if the
	global pointers are changed between the alloc and free
*/
struct __declspec(align(16)) PrefixData
{
	/** Function to free the allocation */
	pfnFree freeFn;
	/** The user data at the time of allocation */
	PVOID pData;
#ifdef _DEBUG
	/** The stack at the time of allocation */
	PVOID* ppStack;
#endif
};

#pragma region Debug Routines
#ifdef _DEBUG
#include <set>
#include <memory>

static HANDLE g_hProcessHeap = GetProcessHeap();
static PVOID DefAlloc(PVOID pHandle, SIZE_T bytes);
static void DefFree(PVOID pHandle, PVOID pData);

/** Allocator which directly uses the Win32 heap functions
	Inherits from allocator for those members which 
	don't need to be changed. Yes, I know this is technically bad
	but its only debugging code.
	\tparam Type The type to allocate memory for
*/
template<class Type>
struct StraightHeapAlloc : public std::allocator<Type>
{
	/** Base typedef for clarity */
	typedef std::allocator<Type> BaseAlloc;
	/** Pointer typedef, also for clarity */
	typedef typename BaseAlloc::pointer pointer;

	/** Does nothing as there's nothing to construct */
	StraightHeapAlloc()
	{}

	/** Constructs an allocator of one type using a different type
		There's still nothing to construct so it does nothing
	*/
	template<class Other>
	StraightHeapAlloc(const StraightHeapAlloc<Other>&)
	{}

	/** Allows the std library to create a copy of this allocator but for a different type
		\tparam Other The new type to allocate
	*/
	template<class Other>
	struct rebind
	{
		/** Typedef to enable the creation of different type allocators */
		typedef StraightHeapAlloc<Other> other;
	};

	/** Allocates count * sizeof(Type) bytes of memory from the process heap
		and returns it 
		\param count The number of Types to allocate
		\returns The allocated memory
	*/
	pointer allocate(size_t count)
	{
		return static_cast<pointer>(DefAlloc(g_hProcessHeap, count * sizeof(Type)));
	}

	/** Allocates count * sizeof(Type) bytes of memory from the process heap
		and returns it 
		\param count The number of Types to allocate
		\returns The allocated memory
	*/
	pointer allocate(size_t count, void*)
	{
		return this->allocate(count);
	}

	/** Returns memory back to the process heap 
		\param pData The memory to free
	*/
	void deallocate(pointer pData, size_t)
	{
		DefFree(g_hProcessHeap, pData);
	}
};

/** Set of allocation pointers typedeffed for clarity */
typedef std::set<PrefixData*, std::less<PrefixData*>, StraightHeapAlloc<PrefixData*> > AllocList;

/** The list of allocations that have yet to be freed */
static AllocList g_outstandingAllocations;
/** The lock required to access g_outstandingAllocations */
static SRWLOCK g_allocListLock;

/** Add an allocation to the outstanding alloc list
	\param pData The allocation to add
*/
static void AddAllocToList(PrefixData* pData)
{
	ExclusiveSRWLock lock(&g_allocListLock);
	g_outstandingAllocations.insert(pData);
}

/** Removes an allocation from the outstanding alloc list
	\param pData The allocation to remove
*/
static void RemoveAllocFromList(PrefixData* pData)
{
	size_t erased = 0;
	{
		ExclusiveSRWLock lock(&g_allocListLock);
		erased = g_outstandingAllocations.erase(pData);
	}
	if(erased == 0)
	{
		ASSERT(FALSE);
	}
}

/** Allocates an array of pointers that are filled with a stack trace
	by our friend RtlCaptureStackBackTrace
	\returns The array of stack frames
*/
static PVOID* AllocStackTrace()
{
	PVOID pTrace[200];
	WORD captured = CaptureStackBackTrace(2, ARRAYSIZE(pTrace), pTrace, NULL);
	PVOID* ppTraces = static_cast<PVOID*>(DefAlloc(g_hProcessHeap, (captured + 1) * sizeof(*pTrace)));
	memcpy(ppTraces, pTrace, captured * sizeof(*pTrace));
	ppTraces[captured] = NULL;
	return ppTraces;
}

/** Frees a stack trace. there's nothing special about them
	so its a simple call to DefFree
	\param ppTrace The memory to free
*/
static void FreeStackTrace(PVOID* ppTrace)
{
	DefFree(g_hProcessHeap, reinterpret_cast<PVOID>(ppTrace));
}

void DisplayOutstandingAllocs()
{
	ExclusiveSRWLock lock(&g_allocListLock);
	AllocList::const_iterator iter = g_outstandingAllocations.begin(), end = g_outstandingAllocations.end();
	UINT index = 1;
	while(iter != end)
	{
		char buffer[300];
		UINT bufferLeft = ARRAYSIZE(buffer);
		PrefixData* pPrefix = *iter;
		int written = StringCchPrintfA(buffer, bufferLeft, "%lu. memory at %#p wasn't freed. Stack Trace:\n", index, pPrefix);
		PVOID* ppStack = pPrefix->ppStack;
		PVOID* ppStackCopy = ppStack;
		UINT stackPos = 1;
		while(*ppStackCopy)
		{
			written += StringCchPrintfA(buffer + written, bufferLeft - written, "%lu. %#p\n\n", stackPos, *ppStackCopy);
			++ppStackCopy;
		}
		FreeStackTrace(ppStack);
		OutputDebugStringA(buffer);
		pPrefix->freeFn(pPrefix->pData, pPrefix);
		++iter;
		++index;
	}
	// if there were any leaks, stop
	if(index != 1)
	{
		ASSERT(FALSE);
	}
}

#else
#define AddAllocToList(x)
#define RemoveAllocFromList(x)
#endif
#pragma endregion

/** The default allocation function

	Requests its memory from the process heap
	\param pHandle The heap handle
	\param bytes How much to allocate
	\returns The memory block
*/
static PVOID DefAlloc(PVOID pHandle, SIZE_T bytes)
{
	return HeapAlloc(pHandle, 0, bytes);
}

/** The default free function

	Returns the memory back to the process heap whence it came
	\param pHandle The heap handle
	\param The memory block to free
*/
static void DefFree(PVOID pHandle, PVOID pData)
{
	HeapFree(pHandle, 0, pData);
}

#pragma region File scope globals
static PVOID g_pUserData;
static pfnAlloc g_AllocatorFn;
static pfnFree g_FreeFn;
static SRWLOCK g_allocLock;
#pragma endregion

// these overloads are documented in stdafx.h
void* __cdecl operator new(size_t bytes)
{
	size_t bytesAndPrefix = bytes + sizeof(PrefixData);
	PVOID pUserData = g_pUserData;

	SharedSRWLock lock(&g_allocLock);
	PrefixData* pData = static_cast<PrefixData*>(g_AllocatorFn(pUserData, bytesAndPrefix));
	pData->freeFn = g_FreeFn;
	pData->pData = pUserData;
#ifdef _DEBUG
	pData->ppStack = AllocStackTrace();
#endif
	AddAllocToList(pData);
	return pData + 1;
}

void* __cdecl operator new[](size_t bytes)
{
	return operator new(bytes);
}

void __cdecl operator delete(void* ptr)
{
	if(ptr)
	{
		SharedSRWLock lock(&g_allocLock);
		PrefixData* pPrefix = static_cast<PrefixData*>(ptr);
		--pPrefix;
		RemoveAllocFromList(pPrefix);
#ifdef _DEBUG
		FreeStackTrace(pPrefix->ppStack);
#endif
		pPrefix->freeFn(pPrefix->pData, pPrefix);
	}
}

void __cdecl operator delete[](void* ptr)
{
	operator delete(ptr);
}

PVOID SetAllocFreeFunctionsInternal(pfnAlloc allocFn, pfnFree freeFn, PVOID pData)
{
	ExclusiveSRWLock lock(&g_allocLock);
	PVOID pPrevUserData = g_pUserData;
	g_pUserData = pData;
	g_AllocatorFn = allocFn;
	g_FreeFn = freeFn;
	return pPrevUserData;
}

BOOL InitialiseAlloc(HMODULE)
{
	InitializeSRWLock(&g_allocLock);
	g_pUserData = GetProcessHeap();
	g_AllocatorFn = &DefAlloc;
	g_FreeFn = &DefFree;
#ifdef _DEBUG
	g_hProcessHeap = GetProcessHeap();
	InitializeSRWLock(&g_allocListLock);
#endif
	return TRUE;
}

void UninitialiseAlloc(HMODULE)
{
	DisplayOutstandingAllocs();
}

/** \endcond */
