/** \file convert.cpp
	\brief Implements the SSE sample conversion functions and the other
	non template conversion related functions. This file has the stack checking
	switch (/GS) turned off, as for some reason VS wants to include one in all
	the SSE functions, none of which use any stack space.
	\cond DEV
*/

#include "stdafx.h"
#include "convert.h"
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <mmreg.h>

/** Contains functions and values used in samples conversion
	These are not to be used directly, instead the relevant functions
	are chosen by DetermineSampleConverter() and called indirectly
	by the ConverterFnPtr variable it returns
*/
namespace detail
{
	/** SSE2 specific functions used in the sample conversion process */
	namespace SSE
	{
		/** Stores premultiplied samples to a memory location

			Converts the premultiplied samples from packed float format
			to packed short format before storing the packed shorts
			into memory. This entire function could be simplified with the _mm_cvtps_pi16 intrinsic
			but that's not available when compiling for X64.

			\param pDstSamples Memory location to store the 4 samples
			\param samples The premultiplied samples in packed float format
		*/
		FORCEINLINE void StoreConvertedSamples(short* pDstSamples, const __m128& samples)
		{
			// convert the floats to longs whose highwords are 0
			__m128i shortRangeLongs = _mm_cvtps_epi32(samples);
			// then shuffle the resulting contents from
			// {sample1, 0, sample2, 0, sample3, 0, sample4, 0}
			// to
			// {sample1, sample2, 0, 0, sample3, 0, sample4, 0}
			__m128i sampleBuffer = _mm_shufflelo_epi16(shortRangeLongs, _MM_SHUFFLE(1, 3, 2, 0));
			// and then to
			// {sample1, sample2, 0, 0, sample3, sample4, 0, 0}
			sampleBuffer = _mm_shufflehi_epi16(sampleBuffer, _MM_SHUFFLE(1, 3, 2, 0));
			// then move samples 3 and 4 next to 1 and 2
			// {sample1, sample2, sample3, sample4, sample3, sample4, 0, 0}
			sampleBuffer = _mm_shuffle_epi32(sampleBuffer, _MM_SHUFFLE(1, 3, 2, 0));
			// then copy the first 4 of them to the output buffer
			_mm_storel_epi64(reinterpret_cast<__m128i*>(pDstSamples), sampleBuffer);
		}

		/** Stores premultiplied samples to a memory location

			Converts the premultiplied samples from packed float format
			to packed long format before storing the packed longs
			into memory.

			\param pDstSamples Memory location to store the 4 samples
			\param samples The premultiplied samples in packed float format
		*/
		FORCEINLINE void StoreConvertedSamples(long* pDstSamples, const __m128& samples)
		{
			__m128i longs = _mm_cvtps_epi32(samples);
			_mm_storeu_si128(reinterpret_cast<__m128i*>(pDstSamples), longs);
		}

		/** Converts samples from packed floats to packed shorts

			Firstly, samples are converted from float to long format.
			The packed longs (whose highwords are all zero) are treated
			as a bunch of eight shorts with contents of {x, 0, x, 0, x, 0, x, 0}.
			The shorts are then shuffled so that the shorts all occupy the first half
			of the register (i.e. the format is {x, x, x, x, 0, 0, 0, 0})
			and are returned to the caller

			\param samples The premultiplied samples in packed float format
		*/
		FORCEINLINE __m128i LongShortToShorts(__m128& samples)
		{
			// convert the floats to longs whose highwords are 0
			__m128i shortRangeLongs = _mm_cvtps_epi32(samples);
			// then shuffle the resulting contents from
			// {sample1, 0, sample2, 0, sample3, 0, sample4, 0}
			// to
			// {sample1, sample2, 0, 0, sample3, 0, sample4, 0}
			__m128i sampleBuffer = _mm_shufflelo_epi16(shortRangeLongs, _MM_SHUFFLE(1, 3, 2, 0));
			// and then to
			// {sample1, sample2, 0, 0, sample3, sample4, 0, 0}
			sampleBuffer = _mm_shufflehi_epi16(sampleBuffer, _MM_SHUFFLE(1, 3, 2, 0));
			// then move samples 3 and 4 next to 1 and 2
			// {sample1, sample2, sample3, sample4, 0, 0, 0, 0}
			return _mm_shuffle_epi32(sampleBuffer, _MM_SHUFFLE(1, 3, 2, 0));
		}

		/** Stores four sets of premultiplied samples to a memory location

			Firstly, four sets of samples are converted from float to short format.
			These four sets of samples are then shuffled from four variables
			containing 4 short values each, to two variables containing 8 shorts each
			(i.e. from four {x, 0, x, 0, x, 0, x, 0} to two {x, x, x, x, x, x, x, x})
			The shuffled variables are then written to memory in one blast.
			There are more straight forward ways of implementing this 
			but despite it's length, this method minimizes loads and stores from memory.

			\param [in, out] pDstSamples The memory location to store the 16 samples, it is updated by the call
			\param samples The four sets of premultiplied samples in packed float format
		*/
		FORCEINLINE void StoreConvertedSamples(short*& pDstSamples, __m128 samples[4])
		{
			__m128i firstSamplesLow = LongShortToShorts(samples[0]);
			__m128i firstSamplesHi = LongShortToShorts(samples[1]);
			__m128i secondSamplesLow = LongShortToShorts(samples[2]);
			__m128i secondSamplesHi = LongShortToShorts(samples[3]);

			// there are no integer shuffling intrinsics that work on two __m128i variables
			// so we cast the registers to double and use the double
			// shuffling instructions (this is strictly a compiler type thing
			// no additional instructions or conversions are performed)
			__m128d samplesLowAsDoubles = _mm_castsi128_pd(firstSamplesLow);
			__m128d samplesHighAsDoubles = _mm_castsi128_pd(firstSamplesHi);

			// shuffle the 4 shorts from firstSamplesHi[0-3] into firstSamplesLow[4-7]
			samplesLowAsDoubles = _mm_shuffle_pd(samplesLowAsDoubles, samplesHighAsDoubles, _MM_SHUFFLE2(0, 0));

			// then do the same for the second 8 samples
			__m128d secondSamplesLowDouble = _mm_castsi128_pd(secondSamplesLow);
			__m128d secondSamplesHiDouble = _mm_castsi128_pd(secondSamplesHi);
			secondSamplesLowDouble = _mm_shuffle_pd(secondSamplesLowDouble, secondSamplesHiDouble, _MM_SHUFFLE2(0, 0));

			// then blast both lots of 8 samples to memory
			_mm_storeu_si128((__m128i*)pDstSamples, _mm_castpd_si128(samplesLowAsDoubles));
			_mm_storeu_si128((__m128i*)(pDstSamples + 8), _mm_castpd_si128(secondSamplesLowDouble));
			pDstSamples += 16;
		}

		/** Stores four sets of premultiplied samples to a memory location

			Four sets of samples are converted from packed float to packed long format
			and then copied to memory.

			\param [in, out] pDstSamples The memory location to store the 16 samples, it is updated by the call
			\param samples The four sets of premultiplied samples in packed float format
		*/
		FORCEINLINE void StoreConvertedSamples(long*& pDstSamples, __m128 samples[4])
		{
			// using an array for these 'longs' variables
			// causes lots of loads and stores to memory
			// this method leaves everything in registers
			// until the copies at the end
			__m128i longs1 = _mm_cvtps_epi32(samples[0]);
			__m128i longs2 = _mm_cvtps_epi32(samples[1]);
			__m128i longs3 = _mm_cvtps_epi32(samples[2]);
			__m128i longs4 = _mm_cvtps_epi32(samples[3]);
			_mm_storeu_si128((__m128i*)pDstSamples, longs1);
			_mm_storeu_si128((__m128i*)(pDstSamples + 4), longs2);
			_mm_storeu_si128((__m128i*)(pDstSamples + 8), longs3);
			_mm_storeu_si128((__m128i*)(pDstSamples + 12), longs4);
			pDstSamples += 16;
		}

		/** Retreives and converts 16 samples

			Loads 16 floats from a memory location into SSE variables and 
			multiplies them all by a scale factor

			\param [in, out] pSrcSamples The memory location containing the raw samples
			\param [out] samplePacks Four SSE variables to store the processed samples in
			\param scaleFactors Contains four copies of the required scale factor
		*/
		FORCEINLINE void Get16ConvertedSamples(const float*& pSrcSamples, __m128 samplePacks[4], const __m128& scaleFactors)
		{
			samplePacks[0] = _mm_load_ps(pSrcSamples);
			samplePacks[1] = _mm_load_ps(pSrcSamples + 4);
			samplePacks[2] = _mm_load_ps(pSrcSamples + 8);
			samplePacks[3] = _mm_load_ps(pSrcSamples + 12);
			samplePacks[0] = _mm_mul_ps(samplePacks[0], scaleFactors);
			samplePacks[1] = _mm_mul_ps(samplePacks[1], scaleFactors);
			samplePacks[2] = _mm_mul_ps(samplePacks[2], scaleFactors);
			samplePacks[3] = _mm_mul_ps(samplePacks[3], scaleFactors);
			pSrcSamples += 16;
		}
	}

	/** Simply copies samples from source to dest without converting them

		\param pSrcArray The (srcChannels * numFrames) samples
		\param numFrames The number of frames of data in pSrcArray
		\param pDstArray Output buffer for the dest type samples
		\param sampleSize The size in bytes of one channel sample
		\param numChannels Number of channels of data in the input buffer
		\returns (numChannels * numFrames) * sampleSize
	*/
	DWORD ConvertCopy(LPCVOID pSrcArray, DWORD numFrames, PVOID pDstArray, WORD sampleSize, WORD numChannels)
	{
		DWORD bytesToCopy = sampleSize * numChannels * numFrames;
		memcpy(pDstArray, pSrcArray, bytesToCopy);
		return bytesToCopy; 
	}

	/** \def X_TO_Y_CHANNEL_Z_HELPER(X, Y, Z)
	
		Defines a three argument function object to assign as the sample conversion function.
		\a X and \a Y are the source and destination C++ sample types respectively.
		\a Z can be Up, Down, or Equal to select the desired channel conversion
		function from source to dest
	*/
#define X_TO_Y_CHANNEL_Z_HELPER(X, Y, Z) \
	converter = boost::bind( \
		&detail::ConvertChannel##Z##<X, Y>, \
		_1, \
		_2, \
		_3, \
		inp.nChannels, \
		out.nChannels \
	);

	/** \def DEFINE_DETERMINE_PCM_CHANNEL_X_HELPER(X)
	
		Defines the implementation of the determination function for integer->integer conversions.
		\a X is the channel conversion type, (either Up, Down, or Equal). This parameter is passed
		unchanged to X_TO_Y_CHANNEL_Z_HELPER as the Z parameter. 
	*/
#define DEFINE_DETERMINE_PCM_CHANNEL_X_HELPER(X) \
	static void DeteminePCMChannel##X##Helper(const WAVEFORMATEX& inp, const WAVEFORMATEX& out, ConverterFnPtr& converter) \
	{ \
		if(inp.wBitsPerSample == 32) \
		{ \
			if(out.wBitsPerSample == 32) \
			{ \
				X_TO_Y_CHANNEL_Z_HELPER(long, long, X); \
			} \
			else if(out.wBitsPerSample == 16) \
			{ \
				X_TO_Y_CHANNEL_Z_HELPER(long, short, X); \
			} \
			else \
			{ \
				ASSERT(out.wBitsPerSample == 8); \
				X_TO_Y_CHANNEL_Z_HELPER(long, char, X); \
			} \
		} \
		else if(inp.wBitsPerSample == 16) \
		{ \
			if(out.wBitsPerSample == 32) \
			{ \
				X_TO_Y_CHANNEL_Z_HELPER(short, long, X); \
			} \
			else if(out.wBitsPerSample == 16) \
			{ \
			 	X_TO_Y_CHANNEL_Z_HELPER(short, short, X); \
			} \
			else \
			{ \
				ASSERT(out.wBitsPerSample == 8); \
				X_TO_Y_CHANNEL_Z_HELPER(short, char, X); \
			} \
		} \
		else \
		{ \
			ASSERT(inp.wBitsPerSample == 8); \
			if(out.wBitsPerSample == 32) \
			{ \
				X_TO_Y_CHANNEL_Z_HELPER(char, long, X); \
			} \
			else if(out.wBitsPerSample == 16) \
			{ \
			 	X_TO_Y_CHANNEL_Z_HELPER(char, short, X); \
			} \
			else \
			{ \
				ASSERT(out.wBitsPerSample == 8); \
				X_TO_Y_CHANNEL_Z_HELPER(char, char, X); \
			} \
		} \
	}

	/** Defines the int to int determination function when
	there are more dest channels than source */
	DEFINE_DETERMINE_PCM_CHANNEL_X_HELPER(Up);
	/** Defines the int to int determination function when
	there are more source channels than dest */
	DEFINE_DETERMINE_PCM_CHANNEL_X_HELPER(Down);
	/** Defines the int to int determination function when
	the number of channels are equal */
	DEFINE_DETERMINE_PCM_CHANNEL_X_HELPER(Equal);

	/** \def DEFINE_DETERMINE_FLOAT_CHANNEL_X_HELPER(X)
	
		Defines the implementation of the determination function for float->integer conversions.
		\a X is the channel conversion type, (either Up, Down, or Equal). This parameter is passed
		unchanged to X_TO_Y_CHANNEL_Z_HELPER as the Z parameter.
	*/
#define DEFINE_DETERMINE_FLOAT_CHANNEL_X_HELPER(X) \
	static void DetermineFloatChannel##X##Helper(const WAVEFORMATEX& inp, const WAVEFORMATEX& out, ConverterFnPtr& converter) \
	{ \
		ASSERT(inp.wBitsPerSample == 32); \
		if(out.wBitsPerSample == 32) \
		{ \
			X_TO_Y_CHANNEL_Z_HELPER(float, long, X); \
		} \
		else if(out.wBitsPerSample == 16) \
		{ \
			X_TO_Y_CHANNEL_Z_HELPER(float, short, X); \
		} \
		else \
		{ \
			ASSERT(out.wBitsPerSample == 8); \
			X_TO_Y_CHANNEL_Z_HELPER(float, char, X); \
		} \
	}

	/** Defines the float to ? determination function when
	there are more dest channels than source */
	DEFINE_DETERMINE_FLOAT_CHANNEL_X_HELPER(Up);
	/** Defines the float to ? determination function when
	there are more source channels than dest ones */
	DEFINE_DETERMINE_FLOAT_CHANNEL_X_HELPER(Down);
	/** Defines the float to ? determination function when
	there number of channels are equal */
	DEFINE_DETERMINE_FLOAT_CHANNEL_X_HELPER(Equal);
}

/** Determines the correct function that'll convert between the sample types

	Picks a suitable function and template parameters that'll
	change samples from the input wave format to the output wave format.

	\param inp The input wave format, that from the audio mixing device
	\param out The output wave format, that to feed to the codec
	\param [out] converter On output, the function object that'll 
	call the actual Convert* function when called
*/
void DetermineSampleConverter(const tWAVEFORMATEX& inp, const tWAVEFORMATEX& out, ConverterFnPtr& converter)
{
	// in the simple case, the samples can just be copied
	if( (inp.wFormatTag == out.wFormatTag) &&
		(inp.nChannels == out.nChannels) &&
		(inp.wBitsPerSample == out.wBitsPerSample)
	)
	{
		converter = boost::bind(
			&detail::ConvertCopy,
			_1,
			_2,
			_3,
			static_cast<WORD>(inp.wBitsPerSample / CHAR_BIT),
			inp.nChannels
		);
		return;
	}
	// There can be 8, 16, or 32 but not 24 because that'd require some extra work
	// (not that any of the encoders seem to support it anyway)
	ASSERT(out.wBitsPerSample != 24);
	// otherwise first check for float->int conversions
	if((inp.wFormatTag == 0 && out.wFormatTag == WAVE_FORMAT_PCM))
	{
		ASSERT(inp.wBitsPerSample == 32);
		// equal channels can use SSE if it's supported
		// as this seems to be the most frequent case
		if(inp.nChannels == out.nChannels)
		{
			if(IsSSE2Supported())
			{
				if(out.wBitsPerSample == 16)
				{
					converter = boost::bind(
						&detail::ConvertChannelEqualSSE<short>,
						_1,
						_2,
						_3,
						inp.nChannels
					);
				}
				else
				{
					// I don't think there are any 24 bits input possibilites
					ASSERT(out.wBitsPerSample == 32);
					converter = boost::bind(
						&detail::ConvertChannelEqualSSE<long>,
						_1,
						_2,
						_3,
						inp.nChannels
					);
				}
			}
			else // non SSE2
			{
				detail::DetermineFloatChannelEqualHelper(inp, out, converter);
			}
		}
		else if(inp.nChannels > out.nChannels)
		{
			detail::DetermineFloatChannelDownHelper(inp, out, converter);
		}
		else
		{
			detail::DetermineFloatChannelUpHelper(inp, out, converter);
		}
	}
	// integer->integer conversions
	if((inp.wFormatTag == WAVE_FORMAT_PCM) && (out.wFormatTag == WAVE_FORMAT_PCM))
	{
		if(inp.nChannels > out.nChannels)
		{
			detail::DeteminePCMChannelUpHelper(inp, out, converter);
		}
		else if(inp.nChannels < out.nChannels)
		{
			detail::DeteminePCMChannelDownHelper(inp, out, converter);
		}
		else
		{
			detail::DeteminePCMChannelEqualHelper(inp, out, converter);
		}
	}
}

/** \endcond */
