/** \file convert.h
	\brief Declares and implements function used for converting between
	the mixer and codec formats
	\cond DEV
*/
#ifndef AUDRECORD_CONVERT_H
#define AUDRECORD_CONVERT_H

#pragma once

#include <boost/function.hpp>
#include "ctnumlimits.h"
#include "misc.h"

/** Function object type for the sample conversion function */
typedef boost::function<DWORD (LPCVOID, DWORD, PVOID)> ConverterFnPtr;

/** Contains functions and values used in samples conversion
	These are not to be used directly, instead the relevant functions
	are chosen by DetermineSampleConverter() and called indirectly
	by the ConverterFnPtr variable it returns
*/
namespace detail
{
	/** Base template of a template "If" copied from C++ Templates: A Complete Guide
		http://books.google.co.uk/books?id=EotSAwuBkJoC&pg=PA309&lpg=PA309&ct=result#PPA309,M1

		\tparam cond A compile time boolean condition
		\tparam TrueArg The resulting nested type if cond is true
		\tparam FalseArg The resulting nested type if cond is false
	*/
	template<bool cond, class TrueArg, class FalseArg>
	struct IfThenElse;

	/** True specialisation of the if base template
		typedefs the resulting type to the passed true type
	*/
	template<class TrueType, class FalseType>
	struct IfThenElse<true, TrueType, FalseType>
	{
		/** The type of the template is TrueType
			when the condition is true*/
		typedef TrueType type;
	};

	/** False specialisation of the if base template
		typedefs the resulting type to the passed true type
	*/
	template<class TrueType, class FalseType>
	struct IfThenElse<false, TrueType, FalseType>
	{
		/** Likewise, when the condition is false
			the type of the template is FalseType
		*/
		typedef FalseType type;
	};

	/** Determines the best type for a scale factor between two sample types 
		and holds the value of SOURCE_MAX / DEST_MAX
		\tparam Source The source samples type
		\tparam Dest Destination sample type
	*/
	template<class Source, class Dest>
	struct ScaleFactor
	{
		/** The scale type, uses float for the type if the source type is smaller than the dest type
			and the source type otherwise
		*/
		typedef typename IfThenElse<(sizeof(Source) < sizeof(Dest)), float, Source>::type ScaleType;
		/** The actual scale value, it is not initialised inline because floats can't be for some reason */
		static const ScaleType value;
	};

	/** The definition of the static const value
		It is not inline with the ScaleFactor struct
		since it can sometimes be float
	*/
	template<class Source, class Dest>
	const typename ScaleFactor<Source, Dest>::ScaleType ScaleFactor<Source, Dest>::value = CTNumLimits<Source>::max / static_cast<ScaleType>(CTNumLimits<Dest>::max);

	/** Specialisation of the ScaleFactor struct when source and dest types
		are equal
	*/
	template<class Source>
	struct ScaleFactor<Source, Source>
	{
		/** Only one choice for the ScaleType when the samples are the same type */
		typedef Source ScaleType;
		/** If the source and dest types are the same, the scale factor is always going to be 1 */
		static const unsigned value = 1;
	};

	/** \def CONVERT_PREAMBLE
		Defines the common declaration used by the conversion functions including
		appropriately casted soucre and destination sample pointers and the caching
		of the scale value in a local variable
	*/
	#define CONVERT_PREAMBLE() \
		const SourceSampleType* pSrcSamples = static_cast<const SourceSampleType*>(pSrcArray); \
		DestSampleType* pDstSamples = static_cast<DestSampleType*>(pDstArray); \
		typedef ScaleFactor<SourceSampleType, DestSampleType> ScaleFactorType; \
		const typename ScaleFactorType::ScaleType scaleFactor = ScaleFactorType::value;

	/** \def CONVERT_RETURN
		Evaluates the number of bytes written to the destination array
		by subtracting the iterating pointer from the base dest pointer
		and returning it
	*/
	#define CONVERT_RETURN() \
		return static_cast<DWORD>(reinterpret_cast<BYTE*>(pDstSamples) - static_cast<BYTE*>(pDstArray));

	/** Simply converts source samples to destination samples

		Loops through (srcChannels * numFrames) source samples, converting them to 
		destination samples by dividing by the scale factor. The unused parameter
		would be for the dstChannels variable, but with an equal number of channels
		this is the same value as the srcChannels variable so it is unnecessary

		\param pSrcArray The (srcChannels * numFrames) source samples of type SourceSampleType
		\param numFrames The number of frames of data in pSrcArray
		\param pDstArray Output buffer for the dest type samples
		\param srcChannels Number of channels of data in the input and output
		\returns Number of bytes written to pDstArray 
	*/
	template<class SourceSampleType, class DestSampleType>
	DWORD ConvertChannelEqual(
		LPCVOID pSrcArray,
		DWORD numFrames,
		PVOID pDstArray,
		WORD srcChannels,
		WORD
	)
	{
		CONVERT_PREAMBLE();
		DWORD frameCount = 0;
		while((frameCount++) < numFrames)
		{
			for(WORD i = 0; i < srcChannels; ++i)
			{
				*(pDstSamples++) = static_cast<DestSampleType>((*(pSrcSamples++)) / scaleFactor);
			}
		}
		CONVERT_RETURN();
	}

	/** Simply converts float samples to destination samples

		Loops through (srcChannels * numFrames) float samples, converting them to 
		destination samples by multiplying by max(DestSampleType). These are multiplied
		rather than divided because	float samples are in the range of [-1,1].

		\param pSrcArray The (srcChannels * numFrames) float samples
		\param numFrames The number of frames of data in pSrcArray
		\param pDstArray Output buffer for the dest type samples
		\param srcChannels Number of channels of data in the input and output
		\returns Number of bytes written to pDstArray 
	*/
	template<class DestSampleType>
	DWORD ConvertChannelEqual(
		LPCVOID pSrcArray,
		DWORD numFrames,
		PVOID pDstArray,
		WORD srcChannels
	)
	{
		const float* pSrcSamples = static_cast<const float*>(pSrcArray);
		DestSampleType* pDstSamples = static_cast<DestSampleType*>(pDstArray);
		static const float scaleFactor = static_cast<float>(CTNumLimits<DestSampleType>::max);
		DWORD frameCount = 0;
		while((frameCount++) < numFrames)
		{
			for(WORD i = 0; i < srcChannels; ++i)
			{
				*(pDstSamples++) = static_cast<DestSampleType>((*(pSrcSamples++)) * scaleFactor);
			}
		}
		CONVERT_RETURN();
	}

	namespace SSE
	{
		void StoreConvertedSamples(short* pDstSamples, const __m128& samples);
		void StoreConvertedSamples(long* pDstSamples, const __m128& samples);
		void StoreConvertedSamples(short*& pDstSamples, __m128 samples[4]);
		void StoreConvertedSamples(long*& pDstSamples, __m128 samples[4]);
		__m128i LongShortToShorts(__m128& samples);
		void Get16ConvertedSamples(const float*& pSrcSamples, __m128 samplePacks[4], const __m128& scaleFactors);

		/** Workhorse for converting 16 float samples at a time using SSE

			Loops around numPacks times getting 16 samples from pSrcSamples pointer,
			converting them to the DestSampleType and storing them into the pDstSamples pointer.

			\tparam DestSampleType The integer type to convert the float samples to
			\param pSrcSamples Pointer to the 16 * numPacks float samples, this pointer is updated during each call
			\param pDstSamples Pointer to a buffer which will contain the 16 converted samples of type DestSampleType,
			this pointer is updated during each call
			\param numPacks Number of 16-float 'packs' to convert
			\param scaleFactors Prefilled SSE type containing 4 copies of the scale factor
		*/
		template<class DestSampleType>
		void Convert16SamplesLoop(const float*& pSrcSamples, DestSampleType*& pDstSamples, DWORD numPacks, const __m128& scaleFactors)
		{
			while(numPacks > 0)
			{
				__m128 samplePacks[4];
				Get16ConvertedSamples(pSrcSamples, samplePacks, scaleFactors);
				StoreConvertedSamples(pDstSamples, samplePacks);
				--numPacks;
			}
		}
	}

	/** Converts float samples to destination samples using SSE

		Converts the source sample floats to their integer destination type (int or short)
		first in groups of 16, then in groups of 4 and then individually for any odd leftovers.
		From testing float->integral with channels equal is the most common case, hence
		why this has an SSE implementation.

		\param pSrcArray The (srcChannels * numFrames) float samples
		\param numFrames The number of frames of data in pSrcArray
		\param pDstArray Output buffer for the dest type samples
		\param srcChannels Number of channels of data in the input and output
		\returns Number of bytes written to pDstArray 
	*/
	template<class DestSampleType>
	DWORD ConvertChannelEqualSSE(
		LPCVOID pSrcArray,
		DWORD numFrames,
		PVOID pDstArray,
		WORD srcChannels
	)
	{
		// assert that the source floats are 16 byte aligned
		ASSERT((PtrToUlong(pSrcArray) & 0xF) == 0);
		const float* pSrcSamples = static_cast<const float*>(pSrcArray);
		DestSampleType* pDstSamples = static_cast<DestSampleType*>(pDstArray);
		const float scaleFactor = static_cast<float>(CTNumLimits<DestSampleType>::max);
		UINT numFloats = numFrames * srcChannels;
		UINT numFloatPacks = numFloats / 16;
		__m128 scaleFactors = _mm_load1_ps(&scaleFactor);

		// multiply floats on the scale of [-1.0 to 1.0] by scaleFactor in 16 float chunks
		SSE::Convert16SamplesLoop(pSrcSamples, pDstSamples, numFloatPacks, scaleFactors);

		// then do any 4 float packs left over
		UINT num4Floats = (numFloats & 0xF) / 4;
		while(num4Floats > 0)
		{
			__m128 procSamples = _mm_load_ps(pSrcSamples);
			procSamples = _mm_mul_ps(procSamples, scaleFactors);
			SSE::StoreConvertedSamples(pDstSamples, procSamples);
			pDstSamples += 4;
			pSrcSamples += 4;
			--num4Floats;
		}
		// convert any odd samples left over
		switch(numFloats & 3)
		{
			case 3:
			{
				float sample = *(pSrcSamples++);
				*(pDstSamples++) = static_cast<DestSampleType>(sample * scaleFactor);
			}
			case 2:
			{
				float sample = *(pSrcSamples++);
				*(pDstSamples++) = static_cast<DestSampleType>(sample * scaleFactor);
			}
			case 1:
			{
				float sample = *(pSrcSamples++);
				*(pDstSamples++) = static_cast<DestSampleType>(sample * scaleFactor);
			}
		}  
		CONVERT_RETURN();
	}

	/** Folds down samples when there are more source channels than destination ones

		For every frame, an average of every (srcChannels / dstChannels) 
		scaled samples are taken. Left over channel samples from the division (i.e. srcChannels % dstChannels)
		are included in the average at the rate of one extra per dstChannel until they've
		all been included. 

		\param pSrcArray The (srcChannels * numFrames) samples
		\param numFrames The number of frames of data in pSrcArray
		\param pDstArray Output buffer for the dest type samples
		\param srcChannels Number of channels of data in the input buffer
		\param dstChannels Number of channels of data in the output buffer
		\returns Number of bytes written to pDstArray 
	*/
	template<class SourceSampleType, class DestSampleType>
	DWORD ConvertChannelDown(
		LPCVOID pSrcArray,
		DWORD numFrames,
		PVOID pDstArray,
		WORD srcChannels,
		WORD dstChannels
	)
	{
		CONVERT_PREAMBLE();
		DWORD frameCount = 0;
		WORD sourcesPerDest = srcChannels / dstChannels;
		const WORD leftOver = srcChannels % dstChannels;
		while((frameCount++) < numFrames)
		{
			WORD leftOverCount = leftOver;
			long thisChannelSample = 0;
			for(WORD i = 0; i < dstChannels; ++i)
			{
				WORD samplesConverted = sourcesPerDest;
				// accumulate converted samples
				for(WORD j = 0; j < sourcesPerDest; ++j)
				{
					thisChannelSample += static_cast<long>((*(pSrcSamples++) / scaleFactor));
				}
				// include the unaccounted for samples/channels at the rate of 
				// one extra per dstChannel until we've used them all
				if(leftOverCount > 0)
				{
					thisChannelSample += static_cast<long>((*(pSrcSamples++) / scaleFactor));
					--leftOverCount;
					++samplesConverted;
				}
				// average it
				thisChannelSample /= samplesConverted;
				// write the average 
				*(pDstSamples++) = static_cast<DestSampleType>(thisChannelSample);
			}
		}
		CONVERT_RETURN();
	}

	/** Replicates samples when there are more destination channels than source ones

		For every frame, one converted source channel is copied to every (dstChannels / srcChannels) 
		channels. Channels unaccounted for by this division (i.e. dstChannels % srcChannels)
		are written at the rate of one per source channel until they've
		all been filled. 

		\param pSrcArray The (srcChannels * numFrames) samples
		\param numFrames The number of frames of data in pSrcArray
		\param pDstArray Output buffer for the dest type samples
		\param srcChannels Number of channels of data in the input buffer
		\param dstChannels Number of channels of data in the output buffer
		\returns Number of bytes written to pDstArray 
	*/
	template<class SourceSampleType, class DestSampleType>
	DWORD ConvertChannelUp(
		LPCVOID pSrcArray,
		DWORD numFrames,
		PVOID pDstArray,
		WORD srcChannels,
		WORD dstChannels
	)
	{
		CONVERT_PREAMBLE();
		DWORD frameCount = 0;
		WORD destsPerSources = dstChannels / srcChannels;
		WORD remainderChans = dstChannels % srcChannels;
		while((frameCount++) < numFrames)
		{
			WORD remainderCount = remainderChans;
			DestSampleType thisChannelSample = 0;
			for(WORD i = 0; i < srcChannels; ++i)
			{
				// replicate converted sample
				thisChannelSample = static_cast<DestSampleType>((*(pSrcSamples++)) / scaleFactor);
				for(WORD j = 0; j < destsPerSources; ++j)
				{
					*(pDstSamples++) = thisChannelSample;
				}
				// plus an extra one to deal with any unaccounted for channels
				if(remainderCount)
				{
					*(pDstSamples++) = thisChannelSample;
					--remainderCount;
				}
			}
		}
		CONVERT_RETURN();
	}

	// documented in the cpp file
	DWORD ConvertCopy(LPCVOID pSrcArray, DWORD numFrames, PVOID pDstArray, WORD sampleSize, WORD numChannels);

#undef CONVERT_PREAMBLE
#undef CONVERT_RETURN
}

// documented in the cpp file
void DetermineSampleConverter(const tWAVEFORMATEX& inp, const tWAVEFORMATEX& out, ConverterFnPtr& converter);

/** \endcond */

#endif
