/** \file messages.cpp
	\brief Implements the message processing functions for
	AudRecord's sampling thread
	\cond DEV
*/

#include "stdafx.h"
#include "messages.h"
#include <mmdeviceapi.h>
#include <audioclient.h>

#pragma region Function table bits
#pragma region Declarations
/** The type of all functions which are called in response to messages */
typedef void DispatchFunction(PVOID pParam);
static DispatchFunction CreateAudioClient;
static DispatchFunction ReleaseAudioClient;
#pragma endregion

/** Dispatch table of functions

	Table is indexed by (msgValue - FIRST_THREAD_MESSAGE)
*/
static DispatchFunction* const g_dispatchFns[] = 
{
	&CreateAudioClient,
	&ReleaseAudioClient
};

// assert that there are as many entries in the message enum than
// there are in the dispatch table
C_ASSERT(ARRAYSIZE(g_dispatchFns) == (LAST_THREAD_MESSAGE - FIRST_THREAD_MESSAGE));

#pragma endregion

/** A blocking equivalent of PostThreadMessage

	Creates a Windows event, posts the message to thread using PostThreadMessage
	and waits for the event to be signalled by the processing thread.
	If the message is out of the range delimited by the ThreadMessage enumeration
	the function fails trivially.

	\param threadId Thread to send the message to
	\param msg Message id to send
	\param pParam Additional message data
	\returns Nonzero if the message was processed, zero if it wasn't
*/
BOOL SendThreadMessage(DWORD threadId, UINT msg, PVOID pParam)
{
	BOOL posted = FALSE;
	if((msg >= FIRST_THREAD_MESSAGE) && (msg < LAST_THREAD_MESSAGE))
	{
		WinType<> hEvent(CreateEvent(NULL, TRUE, FALSE, NULL), &CloseHandle);
		posted = PostThreadMessage(threadId, msg, reinterpret_cast<WPARAM>(*hEvent), reinterpret_cast<LPARAM>(pParam));
		if(posted)
		{
			WaitForSingleObject(*hEvent, INFINITE);
		}
	}
	else
	{
		ASSERT(FALSE);
	}
	return posted;
}

/** Creates an IAudioClient and IAudioCaptureClient for a specified audio endpoint
	
	The function is called in response to a message from the creating thread.
	The interfaces are created here on the dedicated thread because the IAudioClient
	and the IAudioCaptureClient must be released on the thread that created
	them, and this is the only thread that we control.

	\param pParams The parameter data passed in the message, it should be a 
	CreateAudioClientParams structure holding the audio endpoint interface
*/
static void CreateAudioClient(PVOID pParams)
{
	CreateAudioClientParams* pMessageParams = static_cast<CreateAudioClientParams*>(pParams);
	if(pMessageParams)
	{
		HRESULT& hrOut = pMessageParams->hr;
		IMMDevice* pDevice = pMessageParams->pDevice;
		ASSERT(pDevice);
		ComType<IAudioClient> pClientTemp;
		RETURN_ON_FAILED_EXP_VOID(hrOut = pDevice->Activate(IID_IAudioClient, CLSCTX_INPROC_SERVER, NULL, reinterpret_cast<PVOID*>(&pClientTemp)));

		// now we've got the device, grab its wave format
		WAVEFORMATEX* pwfx;
		RETURN_ON_FAILED_EXP_VOID(hrOut = pClientTemp->GetMixFormat(&pwfx));
		WinType<PVOID, void> pwfxWrap(pwfx, &CoTaskMemFree);

		// and do all the initialising
		hrOut = pClientTemp->Initialize(
				AUDCLNT_SHAREMODE_SHARED,
				AUDCLNT_STREAMFLAGS_LOOPBACK | AUDCLNT_STREAMFLAGS_NOPERSIST,
				pMessageParams->bufferDuration,
				0,
				pwfx,
				NULL
			);
		
		if(FAILED(hrOut))
		{
			// this is a recording endpoint, not a rendering one
			if(hrOut == AUDCLNT_E_WRONG_ENDPOINT_TYPE)
			{
				RETURN_ON_FAILED_EXP_VOID(hrOut = pClientTemp->Initialize(
					AUDCLNT_SHAREMODE_SHARED,
					AUDCLNT_STREAMFLAGS_NOPERSIST,
					pMessageParams->bufferDuration,
					0,
					pwfx,
					NULL
				));
			}
			// otherwise some other shenanigans happened
			else RETURN_ON_FAILED_EXP_VOID(("Initialising the Audio Device", hrOut));
		}

		RETURN_ON_FAILED_EXP_VOID(hrOut = pClientTemp->GetService(IID_PPV_ARGS(&pMessageParams->pCaptureClient)));
		// everything succeeded, put the output params in the struct
		pMessageParams->pClient = pClientTemp.Release();
		pMessageParams->pwfx = pwfx;
		pwfxWrap.Release();
	}
	else
	{
		// this should never happen
		ASSERT(FALSE);
	}
}

/** Releases the interfaces created by CreateAudioClient
	
	The function is called in response to a message from the creating thread.
	The interfaces are created here on the dedicated thread because the IAudioClient
	and the IAudioCaptureClient must be released on the thread that created
	them, and this is the only thread that we control.

	\param pParams The parameter data passed in the message, it should be a 
	ReleaseAudioClientParams structure holding the IAudioClient and IAudioCaptureClient
*/
static void ReleaseAudioClient(PVOID pParams)
{
	ReleaseAudioClientParams* pReleaseParams = static_cast<ReleaseAudioClientParams*>(pParams);
	if(pReleaseParams)
	{
		pReleaseParams->pCaptureClient->Release();
		pReleaseParams->pClient->Release();
	}
	else
	{
		// shouldn't happen
		ASSERT(FALSE);
	}
}

/** Responds to a thread message by calling the appropriate callback
	and then signals the event to enable the calling thread to continue.

	If the message is a standard Windows message, Translate/DispatchMessage
	are called instead and no event is set

	\param msg The message to act on
*/
void ActOnMessage(const MSG& msg)
{
	HANDLE hEvent = reinterpret_cast<HANDLE>(msg.wParam);
	PVOID pParams = reinterpret_cast<PVOID>(msg.lParam);
	UINT message = msg.message;
	if(message >= FIRST_THREAD_MESSAGE && message < LAST_THREAD_MESSAGE)
	{
		g_dispatchFns[message - FIRST_THREAD_MESSAGE](pParams);
		SetEvent(hEvent);
	}
	else
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

/** \endcond */
