/** \file misc.cpp
	\brief Implements the miscellaneous functions defined in misc.h
	\cond DEV
*/

#include "stdafx.h"
#include "misc.h"
#include <objbase.h>
#include <mmreg.h>
#include <ctime>

/** Simple check to see whether COM has been initialised for the current thread

	\returns S_OK if it has, CO_E_NOTINITIALIZED if not
*/
HRESULT CheckCOMInitialization()
{
	ULONG_PTR ctx = 0;
	// the ptr returned doesn't have to be released
	// not that MSDN tells you that
	return CoGetContextToken(&ctx);
}

/** Returns the current resolution of the system timer

	The function queries the value set by timeBeginPeriod, 
	in essence it is timeGetPeriod.

	\returns The millisecond resolution of the system timer
*/
DWORD GetCurrentTimerResolution()
{
	// should be NTSTATUS but the typedef is missing, adding a header solely for that seems overkill
	typedef long (NTAPI*pfnNtQueryTimerResolution)(PULONG pMaxRes, PULONG pMinRes, PULONG pCurRes);
	HMODULE hNtDll = GetModuleHandle(L"ntdll.dll");
	pfnNtQueryTimerResolution ntQueryTimerResolution = 
		reinterpret_cast<pfnNtQueryTimerResolution>(GetProcAddress(hNtDll, "NtQueryTimerResolution"));
	DWORD currentMSResolution = 10; // assume this is default
	if(ntQueryTimerResolution)
	{
		ULONG curResIn100NsUnits = 0, minResDummy, maxResDummy;
		ntQueryTimerResolution(&maxResDummy, &minResDummy, &curResIn100NsUnits);
		currentMSResolution = curResIn100NsUnits / REFTIMES_PER_MILLISEC;
	}
	return currentMSResolution;
}

/** Determines the validity and sample type of a WAVEFORMATEX structure

	\param pwfx The structure to examine
	\param [out] pIsFloat If the function returns nonzero, this is set to nonzero
	if the structure describes a format with float samples, zero otherwise.

	\returns Nonzero if the WAVEFORMATEX struct has a WAVE_FORMAT_EXTENSIBLE type
*/
BOOL GetWaveformatExType(const struct tWAVEFORMATEX* pwfx, BOOL* pIsFloat)
{
	BOOL isWfx = (pwfx->wFormatTag == WAVE_FORMAT_EXTENSIBLE);
	if(isWfx)
	{
		WAVEFORMATEXTENSIBLE* pExFormat = CONTAINING_RECORD(pwfx, WAVEFORMATEXTENSIBLE, Format);
		*pIsFloat = (pExFormat->SubFormat == KSDATAFORMAT_SUBTYPE_IEEE_FLOAT);
	}
	else
	{
		ASSERT(pwfx->wFormatTag == WAVE_FORMAT_PCM);
	}
	return isWfx;
}

/** Deterines whether SSE2 support is present

	\returns Nonzero if SSE2 is available, zero if not
*/
BOOL IsSSE2Supported()
{
	int info[4];
	__cpuid(info, 1);
	// check sse2 extension bit
	return (info[3] & (1 << 26));
}

/** Returns the current date and time in the C locale

	Calls time to get current UTC time, localtime to convert that,
	and then the wide strftime using the classic C locale to get the
	date/time string

	\returns The wide result of strftime with the %#c format string
*/
std::wstring GetCurrentDateTime()
{
	time_t timeInt;
	time(&timeInt);
	struct tm curTime;
	localtime_s(&curTime, &timeInt);
	std::wstring dateTimeString(200, 0);

	// explicitly get the string according to the default locale
	// this sis because any other dll or the exe could've changed it
	// for better or worse
	_locale_t cLoc = _create_locale(LC_TIME, "C");
	size_t retChars;
	while((retChars = _wcsftime_l(&dateTimeString[0], dateTimeString.size(), L"%#c", &curTime, cLoc)) == 0)
	{
		dateTimeString.resize(dateTimeString.size() * 2);
	}
	_free_locale(cLoc);
	// The docs seem to imply the return of wcsftime includes the null, it doesn't
	dateTimeString.resize(retChars);
	return dateTimeString;
}

/** Sets the name if the current thread in the Visual Studio debugger

	As an exception is used to set the name, the entire thing is wrapped
	around an IsDebuggerPresent() check. 
	\param pszThreadName The name of the current thread
*/
void SetCurrentThreadName(const char* pszThreadName)
{
	static const DWORD MS_VC_EXCEPTION = 0x406D1388;
#pragma pack(push,8)
	typedef struct tagTHREADNAME_INFO
	{
	   DWORD dwType; // Must be 0x1000.
	   LPCSTR szName; // Pointer to name (in user addr space).
	   DWORD dwThreadID; // Thread ID (-1=caller thread).
	   DWORD dwFlags; // Reserved for future use, must be zero.
	} THREADNAME_INFO;
#pragma pack(pop)
	THREADNAME_INFO info;
	info.dwType = 0x1000;
	info.szName = pszThreadName;
	info.dwThreadID = GetCurrentThreadId();
	info.dwFlags = 0;

	if(IsDebuggerPresent())
	{
		__try
		{
		  RaiseException(MS_VC_EXCEPTION, 0, sizeof(info)/sizeof(ULONG_PTR), (ULONG_PTR*)&info);
		}
		__except(EXCEPTION_EXECUTE_HANDLER)
		{
		}
	}
}

/** \endcond */
