/** \file nullsink.cpp
	\brief Implements the NullSink class
	and the StaticBuffer class returned by NullSink::AllocateDataUnit()
	\cond DEV
*/

#include "stdafx.h"
#include "nullsink.h"

/** Constructs either a normal or realtime NullSink instance 

	\param realTime Whether the sink needs realtime samples for writing or not
*/
NullSink::NullSink(BOOL realTime)
	: realTime(realTime)
{}

/** Returns pointers for implemented interfaces on the object

	Currently supported interfaces are IUnknown and IWMWriterSink
	\param iid The id of the requested interface
	\param [out] ppv A pointer to a pointer that receives the interface pointer or NULL
	\return S_OK on success, E_NOINTERFACE for unsupported interfaces and E_POINTER if ppObj is NULL
*/
STDMETHODIMP NullSink::QueryInterface(REFIID iid, PVOID* ppv)
{
	if(!ppv) return E_POINTER;
	if(iid == IID_IWMWriterSink || iid == IID_IUnknown)
	{
		AddRef();
		*ppv = this;
		return S_OK;
	}
	*ppv = NULL;
	return E_NOINTERFACE;
}

/** Does nothing, the class is meant to be allocated on the stack

	\returns Always 1
*/
STDMETHODIMP_(ULONG) NullSink::AddRef() {return 1;}

/** Does nothing, the class is meant to be allocated on the stack

	\returns Always 1
*/
STDMETHODIMP_(ULONG) NullSink::Release() {return 1;}

/** Does nothing, called when their is a new stream header to write

	\param pHeader Buffer containing a new header to write
	\returns S_OK if pHeader is nonzero, E_POINTER otherwise
*/
STDMETHODIMP NullSink::OnHeader(INSSBuffer* pHeader)
{
	return pHeader ? S_OK : E_POINTER;
}

/** Called by the writer to determine if the sink requires realtime packets

	\param pRealTime Pointer to a bool that receives the 
	realtime value given to the constructor
	\returns S_OK if pRealTime was valid, E_POINTER otherwise
*/
STDMETHODIMP NullSink::IsRealTime(BOOL *pRealTime)
{
	if(!pRealTime) return E_POINTER;
	*pRealTime = realTime;
	return S_OK;
}

/** Does nothing, called when their is a new data to write to the stream

	\param pData Buffer containing data to write
	\returns S_OK if pData is nonzero, E_POINTER otherwise
*/
STDMETHODIMP NullSink::OnDataUnit(INSSBuffer* pData)
{
	return pData ? S_OK : E_POINTER;
}

/** Does nothing, called when all writing has finishe

	\returns Always S_OK
*/
STDMETHODIMP NullSink::OnEndWriting()
{
	return S_OK;
}

/** Simple wrapper class around a vector
	Used by NullSink as a data packet buffer.
	This is not designed to be used on the stack
*/
class StaticBuffer : public INSSBuffer
{
	/** The data buffer that'll never have anything done to it */
	std::vector<BYTE> staticBuffer;
	/** Object reference count */
	ULONG refCount;

	/** Not implemented */
	StaticBuffer(const StaticBuffer&);
	/** Not implemented */
	void operator=(const StaticBuffer&);

public:
	/** Constructs a buffer object of a specified size

		\param dataSize Initial size of the contained buffer
	*/
	StaticBuffer(DWORD dataSize)
		: staticBuffer(dataSize),
		refCount(1)
	{}

	/** Returns pointers for implemented interfaces on the object

		Currently supported interfaces are IUnknown and INSSBuffer
		\param iid The id of the requested interface
		\param [out] ppv A pointer to a pointer that receives the interface pointer or NULL
		\returns S_OK on success, E_NOINTERFACE for unsupported interfaces and E_POINTER if ppObj is NULL
	*/
	STDMETHODIMP QueryInterface(REFIID iid, LPVOID* ppv)
	{
		if(!ppv) return E_POINTER;
		if(iid == IID_INSSBuffer || iid == IID_IUnknown)
		{
			AddRef();
			*ppv = this;
			return S_OK;
		}
		*ppv = NULL;
		return E_NOINTERFACE;
	}

	/** Increments the object's reference count. Not thread safe.

		\returns The new reference count
	*/
	STDMETHODIMP_(ULONG) AddRef()
	{
		return ++refCount;
	}

	/** Decrements the object's reference count. Not thread safe.

		The object is deleted if the ref count drops to zero
		\returns The new reference count
	*/
	STDMETHODIMP_(ULONG) Release()
	{
		ULONG refCountCopy = --refCount;
		if(refCountCopy == 0)
		{
			delete this;
		}
		return refCountCopy;
	}

	// INSSBuffer 
	/** Queries the length of the stored buffer

		\param [out] pLength Pointer that is set to the length of the buffer
		\returns S_OK if the pointer is valid, E_POINTER if it is not
	*/
	STDMETHODIMP GetLength(DWORD* pLength)
	{
		return pLength ? (*pLength = static_cast<DWORD>(staticBuffer.size()), S_OK) : E_POINTER;
	}

	/** Sets the length of the stored buffer

		If length is smaller than the current buffer size, nothing happens.
		Otherwise the buffer is largened to accommodate the request.

		\param length New length of the buffer in bytes
		\returns Always S_OK
	*/
	STDMETHODIMP SetLength(DWORD length)
	{
		if(staticBuffer.size() >= length)
		{
			return S_OK;
		}
		staticBuffer.resize(length);
		return S_OK;
	}

	/** Queries the maximum length of the stored buffer

		\param [out] pLength Pointer that is set to the max length of the buffer
		\returns S_OK if the pointer is valid, E_POINTER if it is not
	*/
	STDMETHODIMP GetMaxLength(DWORD* pLength)
	{
		return pLength ? (*pLength = static_cast<DWORD>(staticBuffer.max_size()), S_OK) : E_POINTER;
	}

	/** Queries for the buffer pointer

		\param [out] ppBuffer Pointer that receives the buffer pointer
		\returns S_OK if the pointer is valid, E_POINTER if it is not
	*/
	STDMETHODIMP GetBuffer(BYTE** ppBuffer)
	{
		return ppBuffer ? (*ppBuffer = &staticBuffer[0], S_OK) : E_POINTER;
	}

	/** Queries for the buffer pointer and buffer length at the same time

		The function is just a combination of GetBuffer() and GetLength()

		\param [out] ppBuffer Pointer that receives the buffer pointer
		\param [out] pLength Pointer that is set to the length of the buffer
		\returns S_OK on success, otherwise an error from GetBuffer() or GetLength
	*/
	STDMETHODIMP GetBufferAndLength(BYTE** ppBuffer, DWORD* pLength)
	{
		HRESULT hr = GetBuffer(ppBuffer);
		if(SUCCEEDED(hr))
		{
			hr = GetLength(pLength);
		}
		return hr;
	}
};

/** Creates a new buffer of the requested size

	\param dataSize Initial size of the memory buffer
	\param [out] ppBuffer Pointer that receives the interface pointer
	of the wrapper around the new buffer
	\returns S_OK on success, E_OUTOFMEMORY of there is no memory left
*/
STDMETHODIMP NullSink::AllocateDataUnit(DWORD dataSize, INSSBuffer** ppBuffer)
{
	if(!ppBuffer) return E_POINTER;
	StaticBuffer* pBuffer = NULL;
	try
	{
		pBuffer = new StaticBuffer(dataSize);
	}
	catch(const std::bad_alloc&)
	{
		return E_OUTOFMEMORY;
	}
	*ppBuffer = pBuffer;
	return S_OK;
}

/** \endcond */
