/** \file AudRecord\stdafx.h
	\brief Main precompiled header, includes project wide header files
	\cond DEV
*/

#ifndef STDAFX_H
#define STDAFX_H

#pragma once

#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0600
#define _DO_NOT_DECLARE_INTERLOCKED_INTRINSICS_IN_MEMORY
#define NOMINMAX

#include "macroutils.h"
#include <intrin.h>
#include <windows.h>
#include "scoped.h"
#include <string>
#include <vector>
#include "ETWEvents.h"

/** \def LOG_TEXT(Function, string, ...)
	Formats \a string and the additional args in a sprintf manner
	and outputs it to the debug stream using ODPrint() and 
	to an ETW event stream using \a Function.
	String must be a narrow string literal, this is asserted by the macro.
*/
#define LOG_TEXT(Function, string, ...) \
	DISABLE_CONST_COND() \
	do \
	{ \
		C_ASSERT(sizeof(*string) == sizeof(char)); \
		char buffer[300]; \
		LogFormat(buffer, __FUNCTION__  " " string "\n", __VA_ARGS__); \
		Function(__FILE__, __LINE__, __FUNCTION__, buffer + ARRAYSIZE(__FUNCTION__) + 1); \
		OutputDebugStringA(buffer); \
	} \
	while(0); \
	END_DISABLE_CONST_COND()

/** \def SEVERE_TEXT(string, ...)
	Formats \a string and the additional args in a sprintf manner
	and outputs it to the ETW event and debug streams using LogSevereEvent()
	and ODPrint(). String must be a narrow string literal, 
	this is asserted by the macro
*/
#define SEVERE_TEXT(string, ...) \
	LOG_TEXT(LogSevereEvent, string, __VA_ARGS__)

/** \def DBG_TEXT(string, ...)
	Equivalent to the SEVERE_TEXT macro except that it logs to the 
	debug channel instead of the severe one.
*/
#define DBG_TEXT(string, ...) \
	LOG_TEXT(LogDebugEvent, string, __VA_ARGS__)

/** \def RETURN_ON_FAILED_EXP(exp)
	Evaluates the expression \a exp which should return a HRESULT.
	If the expression fails (as determined by the FAILED macro) 
	the failure is logged via the DBG_TEXT() macro 
	and the result of the expression returned as the functions
	exit code
*/
#define RETURN_ON_FAILED_EXP(exp) \
	{ \
		HRESULT hr = (exp); \
		if(FAILED(hr)) \
		{ \
			DBG_TEXT(#exp " failed with error %#08x", hr); \
			return hr; \
		} \
	}

/** \def RETURN_ON_FAILED_EXP_VOID(exp)
	Similar to the RETURN_ON_FAILED_EXP() macro
	except for use in void functions where returning
	the HRESULT would cause compile errors
*/
#define RETURN_ON_FAILED_EXP_VOID(exp) \
	{ \
		HRESULT hr = (exp); \
		if(FAILED(hr)) \
		{ \
			DBG_TEXT(#exp " failed with error %#08x", hr); \
			return; \
		} \
	}

/** REFERENCE_TIME time units (100ns) per second */
static const ULONGLONG REFTIMES_PER_SEC = 10000000ULL;
/** REFERENCE_TIME time units (100ns) per millisecond */
static const ULONGLONG REFTIMES_PER_MILLISEC = 10000ULL;
/** The default audio buffer duration is 1 second, expressed in 100ns units */
static const ULONGLONG AUDIO_BUFFER_DURATION = REFTIMES_PER_SEC;
/** Self explanatory, number of milliseconds the audio buffer contains divided by two */
static const DWORD HALF_BUFFER_DURATION_IN_MILLISEC = AUDIO_BUFFER_DURATION / REFTIMES_PER_MILLISEC / 2;

/** Custom object allocator that allocates either with malloc
	or with the user defined allocator
*/
void* __cdecl operator new(size_t bytes);
/** Custom array allocator that allocates either with malloc
	or with the user defined allocator
*/
void* __cdecl operator new[](size_t bytes);
/** Overloaded object deleter that frees either with free
	or with the user defined deleter
*/
void __cdecl operator delete(void* ptr);
/** Overloaded object array deleter that frees either with free
	or with the user defined deleter
*/
void __cdecl operator delete[](void* ptr);

/** \endcond */

#endif
