/** \file wmfunctions.cpp
	\brief Implements that which uses the Windows Media SDK codec, profile,
	and information functions such as enumeration of codecs and their formats and
	matching mixer audio formats to codec inputs.
	\cond DEV
*/

#include "stdafx.h"
#include "wmfunctions.h"
#include "exports.h"
#include <wmsdk.h>
#include <map>
#include "AudRecordInt.h"
#include "misc.h"

/** Temporary holder of a codec configuration
	consisting of an interface and a name
*/
struct CodecConfigTemp
{
	/** The interface containing the config details 
		that can be applied to a IWMWriter profile
	*/
	ComType<IWMStreamConfig> pStreamConfig;
	/** The description of the configuration */
	std::wstring name;

	/** Initialises an object with a name and config interface pointer 
	
		\param pStream The config interface
		\param name Description/name of the config
	*/
	CodecConfigTemp(IWMStreamConfig* pStream, const std::wstring& name)
		: pStreamConfig(pStream),
		name(name)
	{}

	/** Copies an existing object to the members of this one
	
		\param other The object to copy members from
	*/
	CodecConfigTemp(const CodecConfigTemp& other)
		: pStreamConfig(other.pStreamConfig),
		name(other.name)
	{}

	/** Releases the interface pointer from its wrapper
		so it won't be released in the objects destructor
	*/
	void ReleaseInterfaceReference()
	{
		pStreamConfig.Release();
	}
};

typedef std::vector<CodecConfigTemp> CodecConfigList;
typedef CodecConfigList::iterator CodecConfigIter;
typedef CodecConfigList::const_iterator CodecConfigConstIter;
typedef std::map<std::wstring, CodecConfigList> CodecInformation;
typedef CodecInformation::iterator CodecInfoIter;
typedef CodecInformation::const_iterator CodecInfoConstIter;

/** Reduces a map of codecs info and descriptions into a single memory block

	No additional references are taken on any of the interfaces held in ci.
	This is for exception safety in case anything goes wrong either allocating
	or copying the data.

	\param ci The map of codec info to serialize
	\param stringSizes Number of characters including nuls for every
	codec and description name
	\returns Serialized memory block containing a copy of the info in ci
*/
static CodecConfigs* SerializeCodecInfo(CodecInformation& ci, size_t stringSizes)
{
	if(ci.empty())
	{
		return NULL;
	}
	size_t numCodecs = ci.size();
	// start with the mem required by the level structure
	size_t requiredMemory = sizeof(CodecConfigs) + ((numCodecs - 1) * sizeof(Codec*));
	size_t secondLevelStructOffset = requiredMemory;
	
	// then add on the second level
	requiredMemory += (sizeof(Codec) * numCodecs);

	size_t numThirdLevel = 0;
	// then third-level structures 
	for(CodecInfoConstIter iter = ci.begin(), end = ci.end();
		iter != end;
		++iter)
	{
		numThirdLevel += iter->second.size() - 1;
	}

	requiredMemory += (numThirdLevel * sizeof(CodecConfiguration));

	size_t stringOffset = requiredMemory;
	// then the strings
	stringSizes *= sizeof(WCHAR);
	requiredMemory += stringSizes;

	// allocate and start filling in the bits
	BYTE* pStart = new BYTE[requiredMemory];
	CodecConfigs* pTheCodecs = reinterpret_cast<CodecConfigs*>(pStart);
	pTheCodecs->num = numCodecs;

	// ptrs to the relevant parts of the memory
	BYTE* pSecondLevel = pStart + secondLevelStructOffset;
	BYTE* pNextCodecEntry = pSecondLevel;
	WCHAR* pStartOfStrings = reinterpret_cast<WCHAR*>(pStart + stringOffset);

	DWORD codecIndex = 0;
	for(CodecInfoIter iter = ci.begin(), end = ci.end();
		iter != end;
		++iter)
	{
		Codec* pCodec = reinterpret_cast<Codec*>(pNextCodecEntry);
		// link up the pointers
		pTheCodecs->pCodecs[codecIndex++] = pCodec;
		pCodec->numConfigs = iter->second.size();

		// fill in the names
		pCodec->name = pStartOfStrings;
		size_t codecNameLen = iter->first.length();
		StringCchCopy(pStartOfStrings, codecNameLen, iter->first.c_str());
		pStartOfStrings += codecNameLen;

		// then the configs
		CodecConfigList& confList = iter->second;
		CodecConfiguration* pThisConfig = pCodec->configs;
		for(CodecConfigIter configIter = confList.begin(), configEnd = confList.end();
			configIter != configEnd;
			++configIter, ++pThisConfig
		)
		{
			// Not using Release() here to unbing the interface from the ComType
			// if something goes wrong then we'll lose our reference to the object
			// so they are released en masse when this function succeeds
			pThisConfig->pConfig = *configIter->pStreamConfig;
			pThisConfig->pName = pStartOfStrings;
			// we don't need to account for any null here
			// It's already contained in the string's length
			size_t configNameLen = configIter->name.length();
			StringCchCopy(pStartOfStrings, configNameLen, configIter->name.c_str());
			pStartOfStrings += configNameLen;
		}
		pNextCodecEntry = reinterpret_cast<BYTE*>(pThisConfig);
	}
	return pTheCodecs;
}

/** Enumerates the CBR and VBR output formats for a specific codec

	Based on the sample code at
	http://msdn.microsoft.com/en-us/library/dd743771(v=vs.85).aspx

	\param type The WMMEDIATYPE_* media type of the codecs to find, will be WMMEDIATYPE_Audio
	\param pCodec Codec information inteface whose data is to be enummed
	\param codecIndex The codec in pCodec to enum
	\param configList A vector that will have the configurations for this codec added to it
	\param stringSizes Count that is updated with the length of every 
	name and description string encountered.
	\returns S_OK on succes, NS_E_* error on failure
*/
static HRESULT EnumCodecFormats(REFGUID type, IWMCodecInfo3* pCodec, DWORD codecIndex, CodecConfigList& configList, size_t& stringSize)
{
	DWORD vbrEnummed = 0;
	while(vbrEnummed < 2)
	{
		HRESULT hr;
		DWORD numFormats = 0;
		RETURN_ON_FAILED_EXP(pCodec->GetCodecFormatCount(type, codecIndex, &numFormats));
		configList.reserve(numFormats + configList.size());
		for(DWORD i = 0; i < numFormats; ++i)
		{
			IWMStreamConfig* pConfig = NULL;
			DWORD descBufSize = 0;
			pCodec->GetCodecFormatDesc(type, codecIndex, i, NULL, NULL, &descBufSize);

			std::wstring formatDesc(descBufSize, 0);
			if(SUCCEEDED(hr = pCodec->GetCodecFormatDesc(type, codecIndex, i, &pConfig, &formatDesc[0], &descBufSize)))
			{
				// the original call lies and doesn't tell you the actual size
				// required for the string. Filling it in with the second call does
				formatDesc.resize(descBufSize);
				DBG_TEXT("Found config named '%S'", formatDesc.c_str());
				stringSize += descBufSize;
				configList.push_back(CodecConfigTemp(pConfig, formatDesc));
			}
			else
			{
				DBG_TEXT("Failed to get codec format description (codec %lu, format %lu), error %#08x", codecIndex, i, hr);
			}
		}
		// after the first lot of enumming, set the VBR settings so on the next
		// go round, VBR profiles will be enummed
		if((vbrEnummed++) == 0)
		{
			BOOL vbrEnabled = TRUE;
			hr = pCodec->SetCodecEnumerationSetting(
				type,
				codecIndex,
				g_wszVBREnabled,
				WMT_TYPE_BOOL,
				reinterpret_cast<BYTE*>(&vbrEnabled),
				sizeof(vbrEnabled)
			);
			if(hr == NS_E_UNSUPPORTED_PROPERTY)
			{
				DBG_TEXT("Codec doesn't support VBR properties");
				break;
			}
		}
	}
	return S_OK;
}

/** Enumerates all codecs available for use on the system
	and their CBR and VBR configurations.

	Loops through all the codecs getting their name and configurations
	and stores them in a map. The map is then serialised into one big memory block
	by SerializeCodecConfigs(). The references

	\param [out] pConfigs On success, a pointer that is set to a struct
	containing information about the codecs and their info
	\returns S_OK on success, COM or NS_E_* error code on failure
*/
HRESULT GetCodecAndProfileDescriptions(CodecConfigs*& pConfigs)
{
	ComType<IWMProfileManager> pManager;
	RETURN_ON_FAILED_EXP(WMCreateProfileManager(&pManager));
	ComType<IWMCodecInfo3> pCodecInfo;
	RETURN_ON_FAILED_EXP(pManager->QueryInterface(&pCodecInfo));

	REFGUID typeGuid = WMMEDIATYPE_Audio;
    
    // Retrieve the number of supported audio codecs on the system.
	DWORD numCodecs = 0;
    RETURN_ON_FAILED_EXP(pCodecInfo->GetCodecInfoCount(typeGuid, &numCodecs));

	// running total of bytes required to hold all strings encountered
	size_t stringSizes = 0;
	CodecInformation codecInfo;
    // Loop through all the audio codecs.
    for(DWORD dwCodecIndex = 0; dwCodecIndex < numCodecs; ++dwCodecIndex)
    {
		// Get the codec name
		DWORD codecNameSize = 0;
		pCodecInfo->GetCodecName(typeGuid, dwCodecIndex, NULL, &codecNameSize);
		std::wstring codecName(codecNameSize, 0);
		RETURN_ON_FAILED_EXP(pCodecInfo->GetCodecName(typeGuid, dwCodecIndex, &codecName[0], &codecNameSize));
		// codec name size includes the NULL
		stringSizes += codecNameSize;

		DBG_TEXT("Found codec named '%S'", codecName.c_str());

		// then its formats
		std::vector<CodecConfigTemp> confList;
		RETURN_ON_FAILED_EXP(EnumCodecFormats(typeGuid, *pCodecInfo, dwCodecIndex, confList, stringSizes));
		if(!confList.empty())
		{
			codecInfo[codecName].swap(confList);
		}
    }
	pConfigs = SerializeCodecInfo(codecInfo, stringSizes);
	// serialisation completed successfully
	// release the wrappers hold on the interfaces
	// the ref count on all the interfaces is currently 1
	// so if we don't release our hold they'll all be freed at 
	// the end of the function
	if(pConfigs)
	{
		for(CodecInfoIter iter = codecInfo.begin(), end = codecInfo.end();
			iter != end;
			++iter
		)
		{
			std::for_each(
				iter->second.begin(),
				iter->second.end(),
				std::mem_fun_ref(&CodecConfigTemp::ReleaseInterfaceReference)
			);
		}
	}
	return pConfigs ? S_OK : E_OUTOFMEMORY;
}

/** Frees the memory and references associasted with the codecs
	and their configurations

	\returns pCodecConfigs Codec configurations to release
*/
void FreeCodecAndProfileDescriptions(CodecConfigs* pCodecConfigs)
{
	ASSERT(pCodecConfigs);
	DWORD numCodecs = pCodecConfigs->num;
	Codec** pCodecs = pCodecConfigs->pCodecs;
	for(DWORD i = 0; i < numCodecs; ++i)
	{
		Codec* pCodec = pCodecs[i];
		DWORD numConfigs = pCodec->numConfigs;
		CodecConfiguration* pCodecConfig = pCodec->configs, *pEnd = pCodecConfig + numConfigs;
		while(pCodecConfig != pEnd)
		{
			pCodecConfig->pConfig->Release();
			++pCodecConfig;
		}
	}
	delete [] reinterpret_cast<BYTE*>(pCodecConfigs);
}

/** Holds a waveformat and properties that comprise
	a Windows Media writer input format
*/
struct InputFormat
{
	/** The media properties for this format */
	ComType<IWMInputMediaProps> pProps;
	/** Buffer holding the WAVEFORMATEX struct */
	std::vector<BYTE> wfx;

	/** Constructs a new input format from its wave format and stream properties

		\param pProps The interface describing the stream properties of the format
		\param pMedia The media type holding the WAVEFORMATEX to copy into the member buffer
	*/
	InputFormat(const ComType<IWMInputMediaProps>& pProps, const WM_MEDIA_TYPE* pMedia)
		: pProps(pProps),
		wfx(pMedia->pbFormat, pMedia->pbFormat + pMedia->cbFormat)
	{}

	/** Retrieves the wave format for this input format

		\returns The waveformat from the vector
	*/
	const WAVEFORMATEX* WaveFormat() const
	{
		return reinterpret_cast<const WAVEFORMATEX*>(&wfx[0]);
	}

	/** Retrieves the input props for this input format

		\returns The held media properties interface
	*/
	IWMInputMediaProps* Interface() const
	{
		return *pProps;
	}
};

typedef std::vector<InputFormat> FormatVector;

/** Enumerates all input formats available on the writer

	\param pWriter The writer and indirectly the profile whose input possibilities are enumerated
	\param guidSubType The WMMEDIASUBTYPE_* of the media formats to enumerate
	\param inputFormats Collection that will store the enumerated formats
	\returns S_OK on success, a NS_E_* error on failure
*/
static HRESULT GetAllInputFormats(
	IWMWriter* pWriter, 
	GUID guidSubType,
	FormatVector& inputFormats
)
{
	BYTE mediaTypeStack[200];
	ScopedArray<BYTE> mediaTypeHeap;
	BYTE* pBufferToUse = mediaTypeStack;
	DWORD mediaTypeBufferSize = ARRAYSIZE(mediaTypeStack);

	DWORD numFormats = 0;
    // Find the number of formats supported by this input.
	//
	// the dwInput is hardcoded at 0 because audio codecs only 
	// ever have one input
    RETURN_ON_FAILED_EXP(pWriter->GetInputFormatCount(0, &numFormats));
    
	inputFormats.reserve(numFormats);

    // Loop through all of the supported formats.
    for(DWORD formatIndex = 0; formatIndex < numFormats; ++formatIndex)
    {
		HRESULT hr = S_OK;
		ComType<IWMInputMediaProps> pProps;
		DWORD formatSize = 0;
        // Get the input media properties for the input format.
        if(FAILED(hr = pWriter->GetInputFormat(0, formatIndex, &pProps)))
		{
			continue;
		}

        // Get the size of the media type structure.
        if(FAILED(hr = pProps->GetMediaType(NULL, &formatSize)))
		{
			continue;
		}

		if(formatSize > mediaTypeBufferSize)
		{
			ScopedArray<BYTE> pNewBuffer(formatSize);
			mediaTypeHeap.Swap(pNewBuffer);
			mediaTypeBufferSize = formatSize;
		}

		WM_MEDIA_TYPE* pType = reinterpret_cast<WM_MEDIA_TYPE*>(pBufferToUse);
        
        // Get the media type structure.
        hr = pProps->GetMediaType(pType, &mediaTypeBufferSize);
        if(FAILED(hr))
		{
			continue;
		}

		// check if the format matches what is requested
        if(pType->subtype == guidSubType)
        {
			inputFormats.push_back(InputFormat(pProps, pType));			
        }
    }
    return S_OK;
}

/** Finds a matching media properties interface for a specified wave format

	The function loops through all input formats trying to find a match
	between their waveformat and the specified one. A match is defined as
	the same sample rate, channels and bps.

	This and MatchCodecInputFormat() are adapted
	from the WMSDK sample at http://msdn.microsoft.com/en-us/library/dd743773(v=vs.85).aspx

	\param inputFormats The pre-enumerated formats available from the writer
	\param pDesiredFormat The wave format to match
	\param [out] ppProps Pointer to an interface that receives the matching 
	media properties that can be set on the writer
	\returns S_OK if a match was found, NS_E_INVALID_INPUT_FORMAT if one wasn't
*/
static HRESULT FindInputFormat(
	const FormatVector& inputFormats,
	WAVEFORMATEX* pDesiredFormat,
	IWMInputMediaProps** ppProps
)
{
	HRESULT hr = S_OK;
    // Set the ppProps parameter to point to NULL. This will
    //  be used to check the results of the function later.
    *ppProps = NULL;

	FormatVector::const_iterator iter = inputFormats.begin(), end = inputFormats.end();
    
    while(iter != end)
    {
		const WAVEFORMATEX* pTypeWfx = iter->WaveFormat();
		if(
			pTypeWfx->nSamplesPerSec == pDesiredFormat->nSamplesPerSec &&
			pTypeWfx->nChannels == pDesiredFormat->nChannels &&
			pTypeWfx->wBitsPerSample == pDesiredFormat->wBitsPerSample
		)
		{
			BOOL isFloat = FALSE;
			BOOL isRecognizedWaveFormat = GetWaveformatExType(pTypeWfx, &isFloat);
			if(!isRecognizedWaveFormat)
			{
				isRecognizedWaveFormat = (pTypeWfx->wFormatTag == WAVE_FORMAT_PCM);
				isFloat = FALSE;
			}
			// pDesiredFormat->wFormatTag will be 0 for float, and 1 for PCM
			// so always the opposite of the isFloat variable
			if(isRecognizedWaveFormat && (isFloat == !(pDesiredFormat->wFormatTag)))
			{
				pDesiredFormat->wFormatTag = pTypeWfx->wFormatTag;
				*ppProps = iter->Interface();
				(*ppProps)->AddRef();
				break;
			}
		}
		++iter;
    }
	if(!*ppProps)
	{
		// If execution made it to this point, no matching format was found.
		hr = NS_E_INVALID_INPUT_FORMAT;
	}
    return hr;
}

/** Finds the closest codec input wave format for a source wave format

	Enumerates all input formats available on the writer, then tries to match
	the number of channels, sample rate, and bits per sample of the input format.
	If there isn't an exact match, the algorithm to find a match uses the following steps:
		Bits per sample is continually degraded by 8 bits until there's a match
		Sample format is set to integer PCM then step 1 is retried again
		Number of channels is reduced to 2, then step 2 is retried again
		Gives up

	\param pWriter The Windows Media writer whose codec input formats are to be used
	\param sourceFormat Source format to try and match
	\param codecFormat Closest match returned by the function
	\returns S_OK if a match was found, NS_E_INVALID_INPUT_FORMAT if a match couldn't be found or
	a standard COM or NS_E_* error on other failure
*/
HRESULT MatchCodecInputFormat(IWMWriter *pWriter, const WAVEFORMATEX& sourceFormat, WAVEFORMATEX& codecFormat)
{
	FormatVector writerInputs;
	RETURN_ON_FAILED_EXP(GetAllInputFormats(pWriter, WMMEDIASUBTYPE_PCM, writerInputs));
	ComType<IWMInputMediaProps> pProps;
	HRESULT hr = S_OK;
	WAVEFORMATEX codecInputFormat = sourceFormat;
	WORD origBps = codecInputFormat.wBitsPerSample;
	// first try and match bps, format, channels, hz with decreasing bps
	while(
		(codecInputFormat.wBitsPerSample >= 8) &&
		(FAILED(hr = FindInputFormat(writerInputs, &codecInputFormat, &pProps)))
	)
	{
		codecInputFormat.wBitsPerSample -= 8;
	}
	// if the sourceFormat is in floats, try and match bps, channels and hz in integer PCM
	if(FAILED(hr) && (sourceFormat.wFormatTag == 0))
	{
		codecInputFormat.wBitsPerSample = origBps;
		codecInputFormat.wFormatTag = WAVE_FORMAT_PCM;
		while(
			(codecInputFormat.wBitsPerSample >= 8) &&
			(FAILED(hr = FindInputFormat(writerInputs, &codecInputFormat, &pProps)))
		)
		{
			codecInputFormat.wBitsPerSample -= 8;
		}
	}
	// if there are more than 2 channels in the input
	// change down to 2 channels and try and match the bps and hz
	if(FAILED(hr) && (codecInputFormat.nChannels > 2))
	{
		codecInputFormat.wBitsPerSample = origBps;
		codecInputFormat.nChannels = 2;
		while(
			(codecInputFormat.wBitsPerSample >= 8) &&
			(FAILED(hr = FindInputFormat(writerInputs, &codecInputFormat, &pProps)))
		)
		{
			codecInputFormat.wBitsPerSample -= 8;
		}
	}
	// if we've failed here, there is no hope
	if(SUCCEEDED(hr))
	{
		codecFormat = codecInputFormat;
		hr = pWriter->SetInputProps(0, *pProps);
	}
	return hr;
}

/** Creates a IWMWriter interface then creates an AudRecord object

	\param pCodecConfig Initial codec configuration for the recording stream
	\param [out] ppAudRecord Pointer that receives the new interface pointer
	\returns The value returned by AudRecord::Create()
*/
HRESULT CreateAudRecordObject(CodecConfiguration* pCodecConfig, IAudRecord** ppAudRecord)
{
	ComType<IWMWriter> pWriter;
	RETURN_ON_FAILED_EXP(WMCreateWriter(NULL, &pWriter));
	return AudRecord::Create(*pWriter, pCodecConfig, ppAudRecord);
}

/** \endcond */
